/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// qsg.c
// 2000-04-30 QSG COMMON by Maddes

#include "quakedef.h"

// 2000-04-30 QSG COMMON by Maddes
cvar_t	qsg_ssvc_version = {"qsg_ssvc_version", "0", false, true, true, 1, 0, MAX_QSG_VERSION};
cvar_t	qsg_csvc_version = {"qsg_csvc_version", "0", false, false, true, 0, 0, MAX_QSG_VERSION};
cvar_t	qsg_cclc_version = {"qsg_cclc_version", "0", false, false, true, 0, 0, MAX_QSG_VERSION};

cvar_t	qsg_min_version = {"qsg_min_version", "0", false, true};

/*
=====================
QSG_RejectClient

checks client's CMAX version against required QSG version
=====================
*/
// 2000-04-30 QSG HANDSHAKE SRV<->CL/QC<->CL by Maddes
qboolean QSG_RejectClient (void)
{
	// check client against required version
	if (host_client->qsg_cmax < qsg_min_version.value)	// client not compliant
	{
		Con_DPrintf ("Client rejected! His executable/settings doesn't support the required QSG version.\n");

		// tell client and force him to disconnect
		SV_ClientPrintf ("\nRejected! Your executable/settings doesn't support the required QSG version.\n\n");
		cmd_source = src_client;
		Host_Version_f ();
		SV_DropClient (false);
		return true;
	}

	return false;
}


/*
=====================
Host_QSG_Version_f

Command: QSG_VERSION <version>
=====================
*/
// 2000-04-30 QSG HANDSHAKE SRV<->CL/QC<->CL by Maddes
void Host_QSG_Version_f (void)
{
	float	value;

	if (cmd_source == src_command)	// this a client remote only command
	{
		Con_Printf("%s is not valid from the console\n", Cmd_Argv(0));
		return;
	}

	if (Cmd_Argc() < 2)		// not enough arguments
	{
		SV_ClientPrintf("%s: not enough arguments\n", Cmd_Argv(0));
		SV_ClientPrintf("Syntax: %s <version>\n", Cmd_Argv(0));
		return;
	}

	Con_DPrintf("Server received QSG client version %s\n", Cmd_Argv(1));

	// get and check value
	value = Q_atof(Cmd_Argv(1));
	if (value < 0)
	{
		SV_ClientPrintf ("Only positive QSG versions are accepted.\n");
		return;
	}

	// determine and store client QSG versions
	host_client->qsg_cmax = value;
	host_client->qsg_cclc = (value < MAX_QSG_VERSION) ? value : MAX_QSG_VERSION;
	host_client->qsg_csvc = (value < qsg_ssvc_version.value) ? value : qsg_ssvc_version.value;

	// check client against required version
	if (QSG_RejectClient())
	{
		return;
	}

	// tell client the QSG versions
	MSG_WriteByte (&host_client->message, svc_qsgversion);
	MSG_WriteFloat (&host_client->message, qsg_ssvc_version.value);
	MSG_WriteFloat (&host_client->message, host_client->qsg_csvc);
	MSG_WriteFloat (&host_client->message, host_client->qsg_cclc);

	Con_DPrintf("Server sends QSG versions SSVC %1.2f CSVC %1.2f CCLC %1.2f\n", qsg_ssvc_version.value, host_client->qsg_csvc, host_client->qsg_cclc);
}


/*
=====================
Host_QSG_Max_Server_f

Command: QSG_MAX_SERVER <version>
=====================
*/
// 2000-04-30 QSG HANDSHAKE SRV<->QC/SRV<->CL by Maddes
void Host_QSG_Max_Server_f (void)
{
	float	value;

	if (Cmd_Argc() < 2)			// not enough arguments
	{
		Con_Printf("%s: Maximum possible QSG server version is %1.2f", Cmd_Argv(0), qsg_ssvc_version.maxvalue);
		if (qsg_ssvc_version.maxvalue != MAX_QSG_VERSION)
		{
			Con_Printf(", highest maximum is %1.2f", MAX_QSG_VERSION);
		}
		Con_Printf("\nSyntax: %s <version>\n", Cmd_Argv(0));
		return;
	}

	value = Q_atof(Cmd_Argv(1));

	if (value > MAX_QSG_VERSION)
	{
		Con_Printf("%s: QSG server version %1.2f is not supported, highest version is %1.2f\n", Cmd_Argv(0), value, MAX_QSG_VERSION);
		return;
	}

	if (value < 0)
	{
		value = MAX_QSG_VERSION;
	}

	qsg_ssvc_version.maxvalue = value;
	Con_Printf("%s: Maximum possible QSG server version set to %1.2f, change will take effect on next level\n", Cmd_Argv(0), qsg_ssvc_version.maxvalue);
}


/*
=====================
Host_QSG_Max_Client_f

Command: QSG_MAX_CLIENT <version>
=====================
*/
// 2000-04-30 QSG HANDSHAKE SRV<->CL/QC<->CL by Maddes
void Host_QSG_Max_Client_f (void)
{
	float	value;

	if (Cmd_Argc() < 2)			// not enough arguments
	{
		Con_Printf("%s: Maximum possible QSG client version is %1.2f", Cmd_Argv(0), qsg_cclc_version.maxvalue);
		if (qsg_cclc_version.maxvalue != MAX_QSG_VERSION)
		{
			Con_Printf(", highest maximum is %1.2f", MAX_QSG_VERSION);
		}
		Con_Printf("\nSyntax: %s <version>\n", Cmd_Argv(0));
		return;
	}

	value = Q_atof(Cmd_Argv(1));

	if (value > MAX_QSG_VERSION)
	{
		Con_Printf("%s: QSG client version %1.2f is not supported, highest version is %1.2f\n", Cmd_Argv(0), value, MAX_QSG_VERSION);
		return;
	}

	if (value < 0)
	{
		value = MAX_QSG_VERSION;
	}

	qsg_cclc_version.maxvalue = value;
	Con_Printf("%s: Maximum possible QSG client version set to %1.2f, change will take effect on next connect\n", Cmd_Argv(0), qsg_cclc_version.maxvalue);
}


/*
=====================
QSG_Init

Register QSG cvars and commands
=====================
*/
// 2000-04-30 QSG COMMON by Maddes
void QSG_Init (void)
{
	Cvar_RegisterVariable (&qsg_ssvc_version);
	Cvar_RegisterVariable (&qsg_csvc_version);
	Cvar_RegisterVariable (&qsg_cclc_version);

	Cvar_RegisterVariable (&qsg_min_version);

	Cmd_AddCommand ("qsg_max_server", Host_QSG_Max_Server_f);
	Cmd_AddCommand ("qsg_max_client", Host_QSG_Max_Client_f);
	Cmd_AddCommand ("qsg_version", Host_QSG_Version_f);

	sv.qsg_msgwrites = 0;			// 2000-05-02 QSG SVC by Maddes
}


/*
=====================
QSG_InitSVCMsg

Check for modified message and determine conversion tables for clients
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_InitSVCMsg(int dest, int svc_message, int sub_message, client_t *client)
{
	int		i;
	int		search_message;
	msg_lookup_t	*lookup_tab;
	msg_version_t	*version_tab;

	// initialize global data
	sv.qsg_msgserver = NULL;
	sv.qsg_msgwrites = 0;
	for (i=0 ; i<svs.maxclients ; i++)
	{
		svs.clients[i].qsg_msgconversion = NULL;

		if ( (!svs.clients[i].active)			// not active
		|| ((client) && (client != &svs.clients[i])) )	// not selected
		{
			svs.clients[i].qsg_msgignore = true;	// ignore
		}
		else
		{
			svs.clients[i].qsg_msgignore = false;
		}
	}

	// determine which lookup table to use for svc/sub message
	switch(svc_message)
	{
	case svc_temp_entity:
		lookup_tab = te_lookup;
		search_message = sub_message;
		break;

	default:
		lookup_tab = svc_lookup;
		search_message = svc_message;
		break;
	}

	// find svc/sub message in lookup table
	for ( ; lookup_tab->cmd && lookup_tab->cmd != search_message ; lookup_tab++)
	{
	}
	if (!lookup_tab->cmd)			// message not found in lookup table
	{
		return;
	}

	// find server's SSVC in version table
	sv.qsg_msgserver = lookup_tab->version_tab;
	for ( ; sv.qsg_msgserver->qsg_version && sv.qsg_msgserver->qsg_version > qsg_ssvc_version.value ; sv.qsg_msgserver++)
	{
	}
	sv.qsg_msgwrites = sv.qsg_msgserver->numwrites;		// will be decreased through write command

	// find signon's SSVC in version table (last entry 0.00)
	sv.qsg_msgsignon = sv.qsg_msgserver;
	for ( ; sv.qsg_msgsignon->qsg_version ; sv.qsg_msgsignon++)
	{
	}

	// find clients' versions
	for (i=0 ; i<svs.maxclients ; i++)
	{
		if (svs.clients[i].qsg_msgignore)
		{
			continue;		// not active or selected client
		}

		// find client's CSVC in version table
		version_tab = sv.qsg_msgserver;
		for ( ; version_tab->qsg_version && version_tab->qsg_version > svs.clients[i].qsg_csvc ; version_tab++)
		{
		}

		if ( (!version_tab->numwrites)
		|| ((dest == MSG_BROADCAST) && (svs.clients[i].datagram.cursize > MAX_DATAGRAM-version_tab->numbytes)) )
		{
			svs.clients[i].qsg_msgignore = true;
		}
		else
		{
			svs.clients[i].qsg_msgconversion = version_tab->conversion_tab;
		}
	}
}


/*
=====================
PF_QSG_InitSVCMsg

Call QSG_InitSVCMsg from PROGS.DAT/QuakeC
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void PF_QSG_InitSVCMsg(void)
{
	edict_t		*ent;
	int		entnum;
	client_t	*client;

	client = NULL;

	ent = G_EDICT(OFS_PARM3);
	entnum = NUM_FOR_EDICT(ent);
	if (entnum == 0)				// world means all clients
	{
		client = NULL;
	}
	else if (entnum < 1 || entnum > svs.maxclients)
	{
		PR_RunError ("PF_QSG_InitSVCMsg: entity not a client");
	}
	else
	{
		client = &svs.clients[entnum-1];
	}

	QSG_InitSVCMsg(G_FLOAT(OFS_PARM0), G_FLOAT(OFS_PARM1), G_FLOAT(OFS_PARM2), client);
}


/*
=====================
QSG_CheckClient

Checks if client or write command for client should be ignored
=====================
*/
// 2000-05-02 QSG SVC by Maddes
qboolean QSG_CheckClient (client_t *client)
{
	if (client->qsg_msgignore)
	{
		return false;		// client should be ignored (not active, not selected or don't know SVC message)
	}

	// if qsg enhanced message then perform necessary checks
	if (sv.qsg_msgwrites)
	{
		if (!(client->qsg_msgconversion[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites]))
		{
			return false;		// client does not await this data
		}
	}

	return true;
}


/*
=====================
QSG_CheckCounter

Handles write counter
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_CheckCounter (void)
{
	if (sv.qsg_msgwrites)
	{
		sv.qsg_msgwrites--;
	}
}


/*
=====================
QSG_WriteChar
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_WriteChar (int dest, int c, sizebuf_t *sb)
{
	int	i;

	switch (dest)
	{
	case MSG_INIT:
		if (sv.qsg_msgsignon->conversion_tab[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites])
		{
			if (sb)				// special signon
			{
				MSG_WriteChar (sb, c);
			}
			else
			{
				MSG_WriteChar (&sv.signon, c);
			}
		}
		break;

	case MSG_ONE:
	case MSG_ALL:
	case MSG_BROADCAST:
		for (i=0 ; i<svs.maxclients ; i++)
		{
			if (QSG_CheckClient(&svs.clients[i]))
			{
				if (dest == MSG_BROADCAST)	// unreliable
				{
					MSG_WriteChar (&svs.clients[i].datagram, c);
				}
				else if (dest == MSG_ONE && sb)	// special reliable MSG_ONE
				{
					MSG_WriteChar (sb, c);
				}
				else				// reliable MSG_ONE, MSG_ALL
				{
					MSG_WriteChar (&svs.clients[i].message, c);
				}
			}
		}
		break;

	default:
		Host_Error ("QSG_WriteChar: bad destination");
		break;
	}

	QSG_CheckCounter();
}


/*
=====================
QSG_WriteByte
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_WriteByte (int dest, int c, sizebuf_t *sb)
{
	int	i;

	switch (dest)
	{
	case MSG_INIT:
		if (sv.qsg_msgsignon->conversion_tab[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites])
		{
			if (sb)				// special signon
			{
				MSG_WriteByte (sb, c);
			}
			else
			{
				MSG_WriteByte (&sv.signon, c);
			}
		}
		break;

	case MSG_ONE:
	case MSG_ALL:
	case MSG_BROADCAST:
		for (i=0 ; i<svs.maxclients ; i++)
		{
			if (QSG_CheckClient(&svs.clients[i]))
			{
				if (dest == MSG_BROADCAST)	// unreliable
				{
					MSG_WriteByte (&svs.clients[i].datagram, c);
				}
				else if (dest == MSG_ONE && sb)	// special reliable MSG_ONE
				{
					MSG_WriteByte (sb, c);
				}
				else				// reliable MSG_ONE, MSG_ALL
				{
					MSG_WriteByte (&svs.clients[i].message, c);
				}
			}
		}
		break;

	default:
		Host_Error ("QSG_WriteByte: bad destination");
		break;
	}

	QSG_CheckCounter();
}


/*
=====================
QSG_WriteShort
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_WriteShort (int dest, int c, sizebuf_t *sb)
{
	int	i;

	switch (dest)
	{
	case MSG_INIT:
		if (sv.qsg_msgsignon->conversion_tab[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites])
		{
			if (sb)				// special signon
			{
				MSG_WriteShort (sb, c);
			}
			else
			{
				MSG_WriteShort (&sv.signon, c);
			}
		}
		break;

	case MSG_ONE:
	case MSG_ALL:
	case MSG_BROADCAST:
		for (i=0 ; i<svs.maxclients ; i++)
		{
			if (QSG_CheckClient(&svs.clients[i]))
			{
				if (dest == MSG_BROADCAST)	// unreliable
				{
					MSG_WriteShort (&svs.clients[i].datagram, c);
				}
				else if (dest == MSG_ONE && sb)	// special reliable MSG_ONE
				{
					MSG_WriteShort (sb, c);
				}
				else				// reliable MSG_ONE, MSG_ALL
				{
					MSG_WriteShort (&svs.clients[i].message, c);
				}
			}
		}
		break;

	default:
		Host_Error ("QSG_WriteShort: bad destination");
		break;
	}

	QSG_CheckCounter();
}


/*
=====================
QSG_WriteLong
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_WriteLong (int dest, int c, sizebuf_t *sb)
{
	int	i;

	switch (dest)
	{
	case MSG_INIT:
		if (sv.qsg_msgsignon->conversion_tab[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites])
		{
			if (sb)				// special signon
			{
				MSG_WriteLong (sb, c);
			}
			else
			{
				MSG_WriteLong (&sv.signon, c);
			}
		}
		break;

	case MSG_ONE:
	case MSG_ALL:
	case MSG_BROADCAST:
		for (i=0 ; i<svs.maxclients ; i++)
		{
			if (QSG_CheckClient(&svs.clients[i]))
			{
				if (dest == MSG_BROADCAST)	// unreliable
				{
					MSG_WriteLong (&svs.clients[i].datagram, c);
				}
				else if (dest == MSG_ONE && sb)	// special reliable MSG_ONE
				{
					MSG_WriteLong (sb, c);
				}
				else				// reliable MSG_ONE, MSG_ALL
				{
					MSG_WriteLong (&svs.clients[i].message, c);
				}
			}
		}
		break;

	default:
		Host_Error ("QSG_WriteLong: bad destination");
		break;
	}

	QSG_CheckCounter();
}


/*
=====================
QSG_WriteFloat
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_WriteFloat (int dest, float f, sizebuf_t *sb)
{
	int	i;

	switch (dest)
	{
	case MSG_INIT:
		if (sv.qsg_msgsignon->conversion_tab[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites])
		{
			if (sb)				// special signon
			{
				MSG_WriteFloat(sb, f);
			}
			else
			{
				MSG_WriteFloat (&sv.signon, f);
			}
		}
		break;

	case MSG_ONE:
	case MSG_ALL:
	case MSG_BROADCAST:
		for (i=0 ; i<svs.maxclients ; i++)
		{
			if (QSG_CheckClient(&svs.clients[i]))
			{
				if (dest == MSG_BROADCAST)	// unreliable
				{
					MSG_WriteFloat (&svs.clients[i].datagram, f);
				}
				else if (dest == MSG_ONE && sb)	// special reliable MSG_ONE
				{
					MSG_WriteFloat (sb, f);
				}
				else				// reliable MSG_ONE, MSG_ALL
				{
					MSG_WriteFloat (&svs.clients[i].message, f);
				}
			}
		}
		break;

	default:
		Host_Error ("QSG_WriteFloat: bad destination");
		break;
	}

	QSG_CheckCounter();
}


/*
=====================
QSG_WriteString
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_WriteString (int dest, char *s, sizebuf_t *sb)
{
	int	i;

	switch (dest)
	{
	case MSG_INIT:
		if (sv.qsg_msgsignon->conversion_tab[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites])
		{
			if (sb)				// special signon
			{
				MSG_WriteString(sb, s);
			}
			else
			{
				MSG_WriteString (&sv.signon, s);
			}
		}
		break;

	case MSG_ONE:
	case MSG_ALL:
	case MSG_BROADCAST:
		for (i=0 ; i<svs.maxclients ; i++)
		{
			if (QSG_CheckClient(&svs.clients[i]))
			{
				if (dest == MSG_BROADCAST)	// unreliable
				{
					MSG_WriteString (&svs.clients[i].datagram, s);
				}
				else if (dest == MSG_ONE && sb)	// special reliable MSG_ONE
				{
					MSG_WriteString (sb, s);
				}
				else				// reliable MSG_ONE, MSG_ALL
				{
					MSG_WriteString (&svs.clients[i].message, s);
				}
			}
		}
		break;

	default:
		Host_Error ("QSG_WriteString: bad destination");
		break;
	}

	QSG_CheckCounter();
}


/*
=====================
QSG_WriteCoord
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_WriteCoord (int dest, float f, sizebuf_t *sb)
{
	int	i;

	switch (dest)
	{
	case MSG_INIT:
		if (sv.qsg_msgsignon->conversion_tab[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites])
		{
			if (sb)				// special signon
			{
				MSG_WriteCoord(sb, f);
			}
			else
			{
				MSG_WriteCoord (&sv.signon, f);
			}
		}
		break;

	case MSG_ONE:
	case MSG_ALL:
	case MSG_BROADCAST:
		for (i=0 ; i<svs.maxclients ; i++)
		{
			if (QSG_CheckClient(&svs.clients[i]))
			{
				if (dest == MSG_BROADCAST)	// unreliable
				{
					MSG_WriteCoord (&svs.clients[i].datagram, f);
				}
				else if (dest == MSG_ONE && sb)	// special reliable MSG_ONE
				{
					MSG_WriteCoord (sb, f);
				}
				else				// reliable MSG_ONE, MSG_ALL
				{
					MSG_WriteCoord (&svs.clients[i].message, f);
				}
			}
		}
		break;

	default:
		Host_Error ("QSG_WriteCoord: bad destination");
		break;
	}

	QSG_CheckCounter();
}


/*
=====================
QSG_WriteAngle
=====================
*/
// 2000-05-02 QSG SVC by Maddes
void QSG_WriteAngle (int dest, float f, sizebuf_t *sb)
{
	int	i;

	switch (dest)
	{
	case MSG_INIT:
		if (sv.qsg_msgsignon->conversion_tab[sv.qsg_msgserver->numwrites-sv.qsg_msgwrites])
		{
			if (sb)				// special signon
			{
				MSG_WriteAngle(sb, f);
			}
			else
			{
				MSG_WriteAngle (&sv.signon, f);
			}
		}
		break;

	case MSG_ONE:
	case MSG_ALL:
	case MSG_BROADCAST:
		for (i=0 ; i<svs.maxclients ; i++)
		{
			if (QSG_CheckClient(&svs.clients[i]))
			{
				if (dest == MSG_BROADCAST)	// unreliable
				{
					MSG_WriteAngle (&svs.clients[i].datagram, f);
				}
				else if (dest == MSG_ONE && sb)	// special reliable MSG_ONE
				{
					MSG_WriteAngle (sb, f);
				}
				else				// reliable MSG_ONE, MSG_ALL
				{
					MSG_WriteAngle (&svs.clients[i].message, f);
				}
			}
		}
		break;

	default:
		Host_Error ("QSG_WriteAngle: bad destination");
		break;
	}

	QSG_CheckCounter();
}
